use crate::prelude::*;

use anyhow::Result;
use asset::*;
use assetpath::AssetPath;

use std::collections::HashMap;
use std::sync::{Arc, RwLock};
use std::thread;

pub struct AssetManager {
    engine_settings: Arc<EngineSettings>,

    gltf_file_list: Vec<String>,

    gltf_file_map: HashMap<String, AssetPath>,
    glb_file_map: HashMap<String, AssetPath>,

    loaded_gltf_files: Arc<RwLock<HashMap<String, Asset>>>,
    queued_for_async_load: Arc<RwLock<Vec<String>>>,
}

impl AssetManager {
    /// Constructor for `AssetManager`, creates lists of '.gltf' and '.ent' files
    pub(crate) fn new(engine_settings: Arc<EngineSettings>) -> Result<AssetManager> {
        let mut asset_manager = AssetManager {
            engine_settings,

            gltf_file_list: Vec::new(),
            gltf_file_map: HashMap::new(),
            glb_file_map: HashMap::new(),

            loaded_gltf_files: Arc::new(RwLock::new(HashMap::new())),
            queued_for_async_load: Arc::new(RwLock::new(Vec::new())),
        };

        asset_manager.update_files()?;

        Ok(asset_manager)
    }

    #[inline]
    fn find_gltf_file(&self, name: &str) -> Option<&AssetPath> {
        match self.gltf_file_map.get(name) {
            Some(file) => Some(file),
            None => self.glb_file_map.get(name),
        }
    }

    pub fn insert_cached_assets(&self, cache: &AssetCache) -> Result<()> {
        assert!(cache.finished(), "has not finished loading its data");

        let cached_data = cache.get_all()?;
        let mut loaded_gltf_files_lock = self.loaded_gltf_files.write().unwrap();

        for (name, asset) in cached_data.iter() {
            // insert if not present
            if loaded_gltf_files_lock.get(name).is_none() {
                loaded_gltf_files_lock.insert(name.clone(), asset.clone());
            }
        }

        Ok(())
    }

    pub fn create_cache_with<'a, 'b>(
        &mut self,
        context: &Arc<Context>,
        asset_names: impl IntoIterator<Item = &'b &'b str> + Clone,
    ) -> Result<AssetCache> {
        let mut assets = Vec::new();
        let mut names: Vec<(AssetPath, String)> = Vec::new();

        for name in asset_names.into_iter() {
            let name = name.to_string();
            let loaded_gltf_files_lock = self.loaded_gltf_files.read().unwrap();

            match loaded_gltf_files_lock.get(&name) {
                Some(asset) => assets.push((name.clone(), asset.clone())),
                None => names.push((
                    self.find_gltf_file(&name)
                        .ok_or_else(|| {
                            anyhow::Error::msg(format!(
                                "Could not find requested glTF file {}",
                                name
                            ))
                        })?
                        .clone(),
                    name.clone(),
                )),
            }
        }

        AssetCache::new(context, &self.engine_settings, assets, names)
    }

    /// Clears all loaded files
    pub fn clear(&self) -> Result<()> {
        let mut loaded_files_lock = self.loaded_gltf_files.write().unwrap();

        loaded_files_lock.clear();

        Ok(())
    }

    /// Returns all found gltf files
    pub fn gltf_files(&self) -> &Vec<String> {
        &self.gltf_file_list
    }

    /// Updates the lists of all '.gltf' and '.ent' files
    pub fn update_files(&mut self) -> Result<()> {
        let (mut gltf_file_list, gltf_file_map) =
            super::search_file_names(self.engine_settings.gltf_directory(), ".gltf")?;
        let (mut glb_file_list, glb_file_map) =
            super::search_file_names(self.engine_settings.gltf_directory(), ".glb")?;

        gltf_file_list.append(&mut glb_file_list);

        self.gltf_file_list = gltf_file_list;
        self.gltf_file_map = gltf_file_map;
        self.glb_file_map = glb_file_map;

        Ok(())
    }

    /// Creates an empty `EditableEntity`
    pub fn load_asset_file(&mut self, world: &mut World, asset_file: &str) -> Result<EntityObject> {
        // load asset
        let asset = self.load_gltf_asset(world.resources.get::<Context>(), asset_file, true)?;

        let mut entity_object = world.new_entity();

        entity_object.gltf_file = Some(asset.gltf_file);

        let mut bb: BoundingBox = asset.bounding_box.into();

        if !bb.is_zero() {
            bb.check_dimensions(0.25);

            entity_object.insert_component(bb);
        }

        let draw = Draw::new(asset.meshes);

        if !asset.animations.is_empty() {
            entity_object.insert_component(Animation::new(asset.animations, &draw)?);
        }

        entity_object.insert_component(draw);

        Ok(entity_object)
    }

    /// Creates an empty `EditableEntity`
    pub fn load_asset_file_absolute(
        &mut self,
        world: &mut World,
        asset_path: &str,
    ) -> Result<EntityObject> {
        // load asset
        let gltf_asset = GltfAsset::new(&AssetPath::from(("", asset_path)))?;

        let context = world.resources.get::<Context>();

        let asset = Asset::new(
            context.device(),
            context.queue(),
            self.engine_settings.graphics_info()?.render_type,
            gltf_asset,
            asset_path,
        )?;

        let mut entity_object = world.new_entity();

        entity_object.gltf_file = Some(asset.gltf_file);

        let mut bb: BoundingBox = asset.bounding_box.into();

        if !bb.is_zero() {
            bb.check_dimensions(0.25);

            entity_object.insert_component(bb);
        }

        let draw = Draw::new(asset.meshes);

        if !asset.animations.is_empty() {
            entity_object.insert_component(Animation::new(asset.animations, &draw)?);
        }

        entity_object.insert_component(draw);

        Ok(entity_object)
    }

    pub fn async_load_gltf_file<'a, 'b>(
        &self,
        context: &Arc<Context>,
        files: impl IntoIterator<Item = &'b &'b str>,
    ) -> Result<()> {
        let mut asset_files: Vec<(AssetPath, String)> = Vec::new();

        {
            let gltf_files_lock = self.loaded_gltf_files.read().unwrap();

            for gltf_file in files.into_iter() {
                if let Some(path) = self.find_gltf_file(*gltf_file) {
                    // check that asset isn't already loaded
                    if gltf_files_lock.get(*gltf_file).is_none() {
                        asset_files.push((path.clone(), gltf_file.to_string()));
                    }
                }
            }
        }

        // naively spawn for every file a thread
        // TODO: spawn thread dependent on the cpu core count
        for (path, file_name) in asset_files {
            let device = context.device().clone();
            let queue = context.queue().clone();
            let engine_settings = self.engine_settings.clone();

            let loaded_gltf_files = self.loaded_gltf_files.clone();
            let queued_for_async_load = self.queued_for_async_load.clone();

            thread::spawn(move || -> Result<()> {
                if cfg!(debug_assertions) {
                    println!("AssetManager: started thread for gltf file: {}", &path);
                }

                let gltf_asset = GltfAsset::new(&path)?;

                let asset = Asset::new(
                    &device,
                    &queue,
                    engine_settings.graphics_info()?.render_type,
                    gltf_asset,
                    &file_name,
                )?;

                {
                    let mut gltf_files_lock = loaded_gltf_files.write().unwrap();
                    gltf_files_lock.insert(file_name.clone(), asset);
                }

                {
                    let mut queued_lock = queued_for_async_load.write().unwrap();
                    if let Some(index) = queued_lock.iter().position(|name| **name == file_name) {
                        queued_lock.remove(index);
                    }
                }

                if cfg!(debug_assertions) {
                    println!("AssetManager: finished loading gltf file: {}", &path);
                }

                Ok(())
            });
        }

        Ok(())
    }

    /// checks if the '.gltf' file with given name is already loaded, if not, loads it
    /// add - if true, adds the asset to the internal map
    fn load_gltf_asset(&mut self, context: &Context, gltf_file: &str, add: bool) -> Result<Asset> {
        loop {
            let queued_lock = self.queued_for_async_load.read().unwrap();

            // loop until asset is loaded
            if queued_lock.contains(&gltf_file.to_string()) {
                continue;
            } else {
                let mut loaded_gltf_files_lock = self.loaded_gltf_files.write().unwrap();

                match loaded_gltf_files_lock.get(gltf_file) {
                    Some(asset) => return Ok(asset.clone()),
                    None => match self.find_gltf_file(gltf_file) {
                        Some(path) => {
                            let gltf_asset = GltfAsset::new(path)?;

                            let asset = Asset::new(
                                context.device(),
                                context.queue(),
                                self.engine_settings.graphics_info()?.render_type,
                                gltf_asset,
                                gltf_file,
                            )?;

                            if add {
                                loaded_gltf_files_lock.insert(gltf_file.to_string(), asset.clone());
                            }

                            return Ok(asset);
                        }
                        None => {
                            return Err(anyhow::Error::msg(format!(
                                "Could not find requested glTF file {}",
                                gltf_file
                            )));
                        }
                    },
                }
            }
        }
    }
}
