use crate::prelude::*;

use super::RasterizerLightInterface;

use anyhow::Result;
use utilities::{
    impl_reprc,
    prelude::cgmath::{vec3, Matrix4},
};

use std::sync::{Arc, Mutex};

impl_reprc!(
    pub struct ExtendedPointLightData {
        data: PointLightData,
        #[assume_reprc]
        view: Matrix4<f32>,
    }
);

pub struct PointLight {
    light_data: Light,
    light_info_buffer: Arc<Buffer<ExtendedPointLightData>>,

    view: View,

    dummy_map: Arc<Image>,
}

impl RasterizerLightInterface<ExtendedPointLightData> for PointLight {
    fn new(
        light: Light,
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        _image_count: usize,
        shadow_image_size: u32,
    ) -> Result<Self> {
        let mut view = View::new(device, perspective(45.0, 1.0, 0.1, 1000000.0), 45.0, true)?;

        let center = vec3(0.0, 0.0, 0.0);

        view.camera_mut().set_center(center);
        view.camera_mut()
            .set_eye_offset(-light.direction() * light.power());
        view.update_buffer()?;

        // create buffer with light info
        let light_info_buffer = {
            let view_matrix = view.camera_mut().view_matrix();

            Buffer::builder()
                .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
                .set_memory_usage(MemoryUsage::CpuOnly)
                .set_data(&[ExtendedPointLightData {
                    data: PointLightData::from(&light),
                    view: view.proj() * view_matrix,
                }])
                .build(device.clone())?
        };

        let image = Image::empty(
            shadow_image_size,
            shadow_image_size,
            VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
            VK_SAMPLE_COUNT_1_BIT,
        )
        .format(VK_FORMAT_R32G32B32A32_SFLOAT)
        .attach_sampler(Sampler::nearest_sampler().build(device)?)
        .build(device, queue)?;

        Ok(Self {
            light_data: light,
            light_info_buffer,

            view,

            dummy_map: image,
        })
    }

    fn shadow_map(&self) -> &Arc<Image> {
        &self.dummy_map
    }

    fn buffer(&self) -> &Arc<Buffer<ExtendedPointLightData>> {
        &self.light_info_buffer
    }

    fn invalidate(&mut self) {}

    fn render(
        &mut self,
        _content: &Vec<&EntityObject>,
        _buffer_recorder: &mut CommandBufferRecorder<'_>,
        _index: usize,
    ) -> Result<()> {
        // TODO

        Ok(())
    }

    fn set_center(&mut self, center: cgmath::Vector3<f32>) -> Result<()> {
        if self.view.camera().center() != center {
            self.view.camera_mut().set_center(center);

            {
                let mut mapped = self.light_info_buffer.map_complete()?;

                let view_matrix = self.view.camera_mut().view_matrix();
                mapped[0].view = self.view.proj() * view_matrix;
            }

            self.view.update_buffer()?;
        }

        Ok(())
    }

    fn eq(&self, light: &Light) -> bool {
        self.light_data.eq(light)
    }
}

impl Drop for PointLight {
    fn drop(&mut self) {
        self.light_data.clear_update_callback();
    }
}
