#version 450

// input parameter
layout (location = 0) in vec4 in_position;

#ifdef NORMALS
	layout (location = 1) in vec3 in_normal;
#endif

#ifdef NORMAL_UVS
	layout (location = 2) in vec4 in_tangent;
#endif

#ifdef COLOR_UVS
	layout (location = 3) in vec2 in_uvs;
#endif

#ifdef COLOR
	layout (location = 3) in vec4 in_color;
#endif

#ifdef BONES
	layout (location = 4) in ivec4 in_bone_ids;
	layout (location = 5) in vec4 in_bone_weights;
#endif

#ifndef ANIMATE_ONLY
	// out parameter
	layout (location = 0) out vec3 out_world_pos;

	#ifndef CUBE_SHADOW
		#ifdef COLOR_UVS
			layout (location = 1) out vec2 out_uv;
		#endif

		#ifdef NORMALS
			#ifdef NORMAL_UVS
				layout (location = 2) out vec3 out_tangent;
			#endif

			layout (location = 3) out vec3 out_normal;
		#endif
	#endif

	// descriptors
	layout (set = 1, binding = 0) uniform CameraProperties
	{
		mat4 view;
		mat4 proj;
		mat4 vp;

		mat4 inv_view;
		mat4 inv_proj;
	} globals;
#endif

#ifdef SHADOWED
	layout (location = 6) out vec4 out_shadow_coord;

	layout (set = 2, binding = 1) uniform DirectionalLight {
        vec4 direction;
        vec3 color;
        float distance;
        mat4 view;
    } directional_light_info;

	const mat4 bias_matrix = mat4 (
		0.5, 0.0, 0.0, 0.0,
		0.0, 0.5, 0.0, 0.0,
		0.0, 0.0, 1.0, 0.0,
		0.5, 0.5, 0.0, 1.0
	);
#endif

layout (set = 0, binding = 0) uniform ModelTransform
{
	mat4 m;
} model;

#ifdef BONES
	layout (set = 0, binding = 3) uniform BoneData
	{
		mat4 mat[64];
	} bones;
#endif

#ifdef ANIMATE_ONLY
	// storage buffer types are always vec4 aligned
	struct OutputType {
		vec4 position;
		vec4 normal;
		vec4 uv;
	};

	// output buffer
	layout (set = 0, binding = 1) buffer OutputSSBO {
		OutputType values[ ];
	} ssbo;
#endif

void main()
{
	mat4 model_matrix = model.m;

	#ifdef BONES
		mat4 animation_matrix = (bones.mat[in_bone_ids[0]]) * in_bone_weights[0] +
								(bones.mat[in_bone_ids[1]]) * in_bone_weights[1] +
								(bones.mat[in_bone_ids[2]]) * in_bone_weights[2] +
								(bones.mat[in_bone_ids[3]]) * in_bone_weights[3];

		model_matrix = model_matrix * animation_matrix;
	#endif

	vec4 model_pos = model_matrix * in_position;

	#ifdef ANIMATE_ONLY
		ssbo.values[gl_VertexIndex].position = model_pos;

		// set normal
		#ifdef NORMALS
			ssbo.values[gl_VertexIndex].normal = model_matrix * in_normal;
		#else
			ssbo.values[gl_VertexIndex].normal = vec4(0.0);
		#endif

		// set color uv
		#if defined(COLOR_UVS) || defined(NORMAL_UVS)
			ssbo.values[gl_VertexIndex].uv = in_uvs;
		#else
			ssbo.values[gl_VertexIndex].uv = vec4(0.0);
		#endif

		// TODO: indices for textures !?!?!?!?
	#else
		vec4 final_position = globals.vp * model_pos;

		gl_Position = final_position;

		out_world_pos = model_pos.xyz;

		#ifndef CUBE_SHADOW
			#ifdef COLOR_UVS
				out_uv = in_uvs;
			#endif

			#ifdef NORMALS
				mat3 normal = transpose(inverse(mat3(model_matrix)));

				out_normal = normal * normalize(in_normal.xyz);

				#ifdef NORMAL_UVS
					out_tangent = normal * normalize(in_tangent.xyz);
				#endif
			#endif
		#endif
	#endif

	#ifdef SHADOWED
		out_shadow_coord = bias_matrix * directional_light_info.view * model_pos;
	#endif
}
