use crate::game::game::Game;

use anyhow::Result;
use engine::prelude::*;

use engine::prelude::gltf_loader::prelude::*;

use cgmath::{vec2, vec3, Matrix4, SquareMatrix};

use std::{
    sync::{Arc, Weak},
    time::Duration,
};

create_settings_section!(
    SlideSettings,
    "Meta",
    {
        duration: f32,
        count: u32,
    }
);

create_settings_container!(
    SlideContainer,
    {
        settings: SlideSettings,
    }
);

pub struct Slide {
    weak_scene: Weak<Scene>,

    name: String,
    start: Duration,
    duration: Duration,
    count: u32,
    single_width: f32,

    vertex_buffer: Arc<Buffer<PositionColorUV>>,
}

impl Slide {
    pub fn from_create_info(info: SlideCreateInfo) -> Result<Self> {
        todo!()
    }

    pub fn new(name: &str, game: &GameHandle, scene: &Arc<Scene>) -> Result<Arc<Entity>> {
        let info =
            SlideContainer::load(&format!("{}/{}/info.conf", game.slides_directory(), name))?
                .settings;

        let image = {
            let mut slide_images = game.slide_images()?;

            let image_path = format!("{}/{}/texture.png", game.slides_directory(), name);

            match slide_images.get(&image_path) {
                Some(image) => image.clone(),
                None => {
                    let image = Image::from_file(&image_path)?
                        .attach_sampler(Sampler::pretty_sampler().build(game.engine().device())?)
                        .build(game.engine().device(), game.engine().queue())?;

                    slide_images.insert(image_path, image.clone());

                    image
                }
            }
        };

        let (vertex_buffer, width_ratio) = {
            let total_width = image.width();

            let single_width = total_width / info.count;
            let height = image.height();

            let aspect_ratio = single_width as f32 / height as f32;
            let half = aspect_ratio / 2.0;
            let width_ratio = single_width as f32 / total_width as f32;

            let data = [
                PositionColorUV::new(vec3(-half, 0.0, 0.0), vec2(0.0, 1.0)),
                PositionColorUV::new(vec3(half, 0.0, 0.0), vec2(width_ratio, 1.0)),
                PositionColorUV::new(vec3(half, 0.0, 1.0), vec2(width_ratio, 0.0)),
                PositionColorUV::new(vec3(half, 0.0, 1.0), vec2(width_ratio, 0.0)),
                PositionColorUV::new(vec3(-half, 0.0, 1.0), vec2(0.0, 0.0)),
                PositionColorUV::new(vec3(-half, 0.0, 0.0), vec2(0.0, 1.0)),
            ];

            let buffer = Buffer::builder()
                .set_usage(VK_BUFFER_USAGE_VERTEX_BUFFER_BIT | VK_BUFFER_USAGE_STORAGE_BUFFER_BIT)
                .set_memory_usage(MemoryUsage::CpuOnly)
                .set_data(&data)
                .build(game.engine().device().clone())?;

            (buffer, width_ratio)
        };

        let width_ratio = 0.0;

        let meshes = {
            let mut mesh = AssetMesh::new(game.engine().device(), scene.scene_type())?;

            mesh.add_primitive(
                vertex_buffer.clone(),
                Some(image),
                None,
                PrimitiveMaterial {
                    emissive_factor: [1.0, 1.0, 1.0],
                    ..PrimitiveMaterial::default()
                },
                false,
            )?;

            vec![mesh]
        };

        let entity_builder = Entity::new_custom(game.engine().context(), meshes);

        let entity_builder = Entity::new_custom();

        let entity = entity_builder.build()?;

        let slide = Slide {
            weak_scene: Arc::downgrade(scene),

            name: name.to_string(),
            start: Duration::default(),
            duration: Duration::from_secs_f32(info.duration),
            count: info.count,
            single_width: width_ratio,

            vertex_buffer,
        };

        entity.components()?.insert(slide);

        Ok(entity)
    }

    pub fn update(&self, now: Duration) -> Result<()> {
        let time_diff = now - self.start;

        let time = time_diff.as_secs_f32() % self.duration.as_secs_f32();
        let time_slice = self.duration.as_secs_f32() / self.count as f32;

        let slice = (time / time_slice).floor();

        let mut mapped = self.vertex_buffer.map_complete()?;

        mapped[0].uv.x = slice * self.single_width;
        mapped[1].uv.x = (slice + 1.0) * self.single_width;
        mapped[2].uv.x = (slice + 1.0) * self.single_width;
        mapped[3].uv.x = (slice + 1.0) * self.single_width;
        mapped[4].uv.x = slice * self.single_width;
        mapped[5].uv.x = slice * self.single_width;

        Ok(())
    }
}

impl EntityComponent for Slide {
    fn enable(&mut self, scene: &Scene) -> Result<()> {
        if let Some(scene) = self.weak_scene.upgrade() {
            self.start = scene.now();
        }

        Ok(())
    }

    fn name(&self) -> &str {
        "Slide"
    }
}

#[derive(Serialize, Deserialize)]
pub struct SlideCreateInfo {
    // TODO
}

impl<'a> ComponentCreateInfo<'a> for SlideCreateInfo {}
