use anyhow::Result;
use cgmath::{Deg, Vector2, Vector3};
use rpg_components::components::abilityloader::AbilityLoader;

use std::collections::HashSet;
use std::time::Duration;

use crate::game::{
    content::{
        abilities::particle_spawn::{ParticleSpawn, ParticleSpawnExecutor},
        prelude::*,
    },
    game::GameHandle,
};
use engine::prelude::*;

pub struct ParticleInfo {
    pub radius: f32,
}

pub struct AreaOfEffect {
    direction: Vector2<f32>,

    position: Vector3<f32>,

    arc: Deg<f32>,
    radius: f32,
    // height: f32,
    ability_info: AbilityLoader,
    base_damage: u32,

    start: Duration,

    owner: Entity,

    game: GameHandle,

    light: Option<Light>,

    proc_slots: Vec<(Duration, HashSet<Entity>)>,

    audio: Option<UnsafeComponentStore<Audio>>,
}

impl AreaOfEffect {
    pub fn post_setup(&mut self, audio: Option<&mut Audio>) {
        self.audio = audio.map(|audio| audio.into());
    }

    #[allow(clippy::too_many_arguments)]
    #[allow(clippy::new_ret_no_self)]
    pub fn new(
        direction: Vector2<f32>,
        position: Vector3<f32>,
        arc: Deg<f32>,
        radius: f32,
        height: f32,
        owner: Entity,
        ability_info: AbilityLoader,
        base_damage: u32,
        game: GameHandle,
    ) -> Result<(EntityObject, Option<ParticleSpawn>)> {
        let mut particle_spawn = None;

        let proc_slots = Self::calculate_proc_slots(
            ability_info.settings.parameter.procedure_count,
            ability_info.settings.parameter.delay.into(),
            ability_info.settings.parameter.duration.into(),
        )?
        .into_iter()
        .map(|time| (time, HashSet::new()))
        .collect();

        let me = Self {
            direction,
            position: position,
            arc,
            radius,
            // height,
            ability_info: ability_info.clone(),

            start: Duration::default(),
            base_damage,

            owner,
            game,

            light: None,

            audio: None,

            proc_slots,
        };

        let mut entity = me.game.upgrade().engine().assets().empty_entity();

        // don't create an extra entity for now if duration is zero or if the ability has no delay
        if me.ability_info.settings.parameter.duration != PersistentDuration::from_secs(0)
            || me.ability_info.settings.parameter.delay != PersistentDuration::from_secs(0)
        {
            entity.insert_component(Draw::new(vec![]));

            if !me.ability_info.settings.parameter.idle_sound.is_empty() {
                let mut audio = Audio::new(me.game.upgrade().engine().context(), None)?;

                audio.add_custom_sound(
                    me.game
                        .upgrade()
                        .build_data_path(&me.ability_info.settings.parameter.idle_sound),
                    "idle_sound",
                )?;

                entity.insert_component(audio);
            };

            if !ability_info.settings.meta.particles.is_empty() {
                particle_spawn = Some(ParticleSpawn {
                    entity: entity.as_entity(),
                    info: ability_info,
                    executor: ParticleSpawnExecutor::AoE,
                });
            }

            Location::new_and_setup(&mut entity)?.set_position(position);

            entity.insert_component(me.create_aoe_arc()?);
            entity.insert_component(me);
            entity.insert_component(BoundingBox {
                min: [-radius, -radius, 0.0],
                max: [radius, radius, height],
            });
        }

        #[cfg(debug_assertions)]
        {
            entity.debug_name = Some("AoE".to_string());
        }

        Ok((entity, particle_spawn))
    }

    pub fn calculate_proc_slots(
        count: u32,
        delay: Duration,
        duration: Duration,
    ) -> Result<Vec<Duration>> {
        if count == 0 {
            return Err(anyhow::Error::msg("Proc Count 0 in AoE is useless"));
        }

        // first proc is at `delay` time
        let mut procs = vec![delay];

        if count == 1 {
            if duration > Duration::from_secs(0) {
                return Err(anyhow::Error::msg(
                    "Undefined when proc should happen if there is only 1 but duration is there",
                ));
            }

            // nothing else needs to be done here
        } else {
            if duration == Duration::from_secs(0) {
                return Err(anyhow::Error::msg("Multiple procs happen at the same time"));
            }

            let time_slice = duration / (count - 1);

            for i in 1..count {
                procs.push(delay + time_slice * i);
            }
        }

        Ok(procs)
    }

    pub fn check_life_time(&mut self, now: Duration) -> bool {
        let duration: Duration = self.ability_info.settings.parameter.duration.into();
        let delay = self.ability_info.settings.parameter.delay.into();

        (self.start + delay + duration) >= now
    }

    pub fn check_entity(&mut self, now: Duration, collider: Entity) -> bool {
        // find next smallest time
        match self
            .proc_slots
            .iter_mut()
            .rev()
            .find(|(time, _)| (self.start + *time) <= now)
        {
            Some((_, entities)) => entities.insert(collider),
            None => false,
        }
    }

    fn create_aoe_arc(&self) -> Result<AreaOfEffectArc> {
        Ok(AreaOfEffectArc::new(
            self.arc,
            self.radius,
            self.base_damage,
            self.ability_info.settings.parameter.damage_type,
            self.owner,
        ))
    }

    #[inline]
    fn create_lighting(
        engine: &Engine,
        ability_info: &AbilityLoader,
        mut position: Vector3<f32>,
        info: &ParticleSystemInfo,
    ) -> Result<Option<Light>> {
        let mut light = if ability_info.settings.parameter.light {
            let mut light = engine.new_point_light()?;

            position.z += 0.5;
            light.set_position(position)?;
            light.set_power(200.0)?;

            Some(light)
        } else {
            None
        };

        if let Some(light) = &mut light {
            let slice: [f32; 3] = info.particles.end_color.into();
            light.set_color(slice.into())?;
        }

        Ok(light)
    }

    pub fn create_particles(
        mut info: ParticleSystemInfo,
        draw: &mut Draw,
        particle_system_vulkan_objects: &ParticleSystemVulkanObjects,
        engine: &Engine,
        particle_info: ParticleInfo,
    ) -> Result<ParticleSystem> {
        info.system.particle_count = (particle_info.radius * 250.0) as usize;
        info.system.random_start_factor = particle_info.radius;

        ParticleSystem::new_with_vk_objects(info, engine, particle_system_vulkan_objects, draw)
    }

    #[inline]
    pub fn create_particle_effect(
        &mut self,
        game: &GameHandle,
        loader: &AbilityLoader,
        draw: &mut Draw,
        particle_system_vulkan_objects: &ParticleSystemVulkanObjects,
    ) -> Result<ParticleSystem> {
        let info = ParticleSystemInfo::load(game.build_data_path(&loader.settings.meta.particles))?;

        if loader.settings.parameter.procedure_count > 1 {
            self.light =
                Self::create_lighting(game.upgrade().engine(), &loader, self.position, &info)?
        }

        Self::create_particles(
            info,
            draw,
            particle_system_vulkan_objects,
            game.upgrade().engine(),
            ParticleInfo {
                radius: self.radius,
            },
        )
    }

    pub fn direction(&self) -> Vector2<f32> {
        self.direction
    }
}

impl EntityComponent for AreaOfEffect {
    fn enable(&mut self, scene: &mut Scene) -> Result<()> {
        if let Some(light) = &self.light {
            scene.add_light(light.clone())?;
        }

        if let Some(audio) = &mut self.audio {
            unsafe { audio.as_mut() }.play("idle_sound", true)?;
        }

        self.start = scene.now();

        Ok(())
    }

    fn disable(&mut self, scene: &mut Scene) -> Result<()> {
        if let Some(light) = &self.light {
            scene.remove_light(light.clone())?;
        }

        if let Some(audio) = &mut self.audio {
            unsafe { audio.as_mut() }.stop()?;
        }

        Ok(())
    }

    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for AreaOfEffect {
    fn debug_name() -> &'static str {
        "AreaOfEffect"
    }
}

#[derive(Serialize, Deserialize)]
pub struct AreaOfEffectCreateInfo {
    direction: Vector2<f32>,
    position: Vector2<f32>,
    arc: f32,
    radius: f32,
    height: f32,
    ability_info: AbilityLoader,
    base_damage: u32,
    owner: Entity,
}

impl<'a> ComponentCreateInfo<'a> for AreaOfEffectCreateInfo {}

#[test]
fn verify_procs() {
    // ------ check if the procs are splitted correctly ------

    let delay_only =
        AreaOfEffect::calculate_proc_slots(1, Duration::from_secs(1), Duration::from_secs(0))
            .unwrap();

    assert_eq!(delay_only, vec![Duration::from_secs(1)]);

    let split_5_without_delay =
        AreaOfEffect::calculate_proc_slots(5, Duration::from_secs(0), Duration::from_secs(1))
            .unwrap();

    assert_eq!(
        split_5_without_delay,
        vec![
            Duration::from_secs(0),
            Duration::from_secs_f32(0.25),
            Duration::from_secs_f32(0.5),
            Duration::from_secs_f32(0.75),
            Duration::from_secs(1),
        ]
    );

    let split_5_with_delay =
        AreaOfEffect::calculate_proc_slots(5, Duration::from_secs(2), Duration::from_secs(1))
            .unwrap();

    assert_eq!(
        split_5_with_delay,
        vec![
            Duration::from_secs(2),
            Duration::from_millis(2250),
            Duration::from_millis(2500),
            Duration::from_millis(2750),
            Duration::from_secs(3),
        ]
    );

    // ------ check error detection ------

    assert!(
        AreaOfEffect::calculate_proc_slots(2, Duration::from_secs(0), Duration::from_secs(0))
            .is_err()
    );
    assert!(
        AreaOfEffect::calculate_proc_slots(0, Duration::from_secs(0), Duration::from_secs(0))
            .is_err()
    );
    assert!(
        AreaOfEffect::calculate_proc_slots(1, Duration::from_secs(0), Duration::from_secs(1))
            .is_err()
    );
}
