use anyhow::Result;
use cgmath::vec3;
use engine::prelude::*;

use std::sync::Arc;

use crate::Game;

use super::health_bar_base::HealthBarBase;

pub struct BossHealthBar {
    gui: Arc<GuiBuilder>,
    hp_bar: Arc<ProgressBar>,

    base: HealthBarBase,
}

impl BossHealthBar {
    pub fn new(game: &Game, boss_name: impl ToString) -> Result<Self> {
        let gui = GuiBuilder::new(
            game.engine().gui_handler(),
            &game.build_data_path("gui/xml/ingame/boss_healthbar.xml"),
        )?;

        let progress_bar: Arc<ProgressBar> = gui.element("health_bar")?;
        progress_bar.set_progress(1.0)?;

        let name: Arc<Label> = gui.element("boss name")?;
        name.set_text(boss_name)?;

        Ok(Self {
            gui,
            hp_bar: progress_bar,

            base: HealthBarBase::new(vec3(0.0, 0.0, 0.0)),
        })
    }

    pub fn update(&mut self, health: u32, hp_progress: f32, should_render: bool) -> Result<()> {
        if self.base.should_enable && !self.base.is_enabled {
            self.base.is_enabled = true;
            self.gui.enable()?;
        }

        match (should_render, self.base.is_enabled) {
            (true, false) | (true, true) => self.gui.enable()?,
            (false, true) => self.gui.disable()?,

            _ => (),
        }

        println!(
            "should render: {}, enabled: {}",
            should_render, self.base.is_enabled
        );

        self.hp_bar.set_progress(hp_progress)?;
        self.hp_bar.set_text(health)?;

        Ok(())
    }
}

impl EntityComponent for BossHealthBar {
    fn enable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.base.should_enable = true;

        Ok(())
    }

    fn disable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.base.should_enable = false;
        self.base.is_enabled = false;

        self.gui.disable()?;

        Ok(())
    }

    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for BossHealthBar {
    fn debug_name() -> &'static str {
        "BossHealthBar"
    }
}
