use anyhow::Result;
use cgmath::Vector3;
use engine::prelude::*;

use std::{convert::TryFrom, sync::Arc};

use crate::Game;

use super::health_bar_base::HealthBarBase;

pub struct NPCHealthBar {
    gui: Arc<GuiBuilder>,
    main_grid: Arc<Grid>,
    hp_bar: Arc<ProgressBar>,

    base: HealthBarBase,
}

impl NPCHealthBar {
    pub fn new(game: &Game, elite: bool, world_offset: Vector3<f32>) -> Result<Self> {
        let gui = GuiBuilder::new(
            game.engine().gui_handler(),
            &game.build_data_path("gui/xml/ingame/npc_healthbar.xml"),
        )?;

        let progress_bar: Arc<ProgressBar> = gui.element("health_bar")?;
        progress_bar.set_progress(1.0)?;

        let health_bar_grid: Arc<Grid> = gui.element("health_bar_grid")?;

        // this is super important to prevent weird behavior
        health_bar_grid.disallow_position_scale()?;

        if elite {
            health_bar_grid
                .set_background((Color::try_from("#FFD700")?, DisplayableFillType::Expand))?;
        }

        Ok(Self {
            gui,
            main_grid: health_bar_grid,
            hp_bar: progress_bar,

            base: HealthBarBase::new(world_offset),
        })
    }

    pub fn update(&mut self, (health, hp_progress): (u32, f32), (x, y): (i32, i32)) -> Result<()> {
        if self.base.should_enable && !self.base.is_enabled {
            self.base.is_enabled = true;
            self.gui.enable()?;
        }

        self.hp_bar.set_progress(hp_progress)?;
        self.hp_bar.set_text(health)?;

        if self.base.center != (x, y) {
            self.base.center = (x, y);

            let (width, height) = self.main_grid.extents();
            self.main_grid
                .set_position(x - (width as i32 / 2), y - (height as i32 / 2))?;
        }

        Ok(())
    }

    pub fn world_offset(&self) -> Vector3<f32> {
        self.base.world_offset
    }
}

impl EntityComponent for NPCHealthBar {
    fn enable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.base.should_enable = true;

        Ok(())
    }

    fn disable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.base.should_enable = false;
        self.base.is_enabled = false;

        self.gui.disable()?;

        Ok(())
    }

    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for NPCHealthBar {
    fn debug_name() -> &'static str {
        "NPCHealthBar"
    }
}
